import json
from openai.types.beta.threads import (
    FileCitationAnnotation,
    FilePathAnnotation,
    ImageFileContentBlock,
    ImageURLContentBlock,
    Message,
    TextContentBlock
)
from openai.types.beta.threads.runs import (
    CodeInterpreterToolCall,
    FunctionToolCall,
    RunStep
)
from foundationallm.models.constants import AgentCapabilityCategories
from foundationallm.models.orchestration import (
    OpenAIFilePathMessageContentItem,
    OpenAIImageFileMessageContentItem,
    OpenAITextMessageContentItem,
    AnalysisResult
)

class OpenAIAssistantsHelpers:

    @staticmethod
    def parse_run_step(run_step: RunStep) -> AnalysisResult:
        """
        Parses a run step from the OpenAI Assistants API.

        Parameters
        ----------
        run_step : RunStep
            The run step to parse.

        Returns
        -------
        AnalysisResult
            The analysis result from the run step.
        OR None
            If the run step does not contain a tool call
            to the code interpreter tool.
        """
        step_details = run_step.step_details
        if step_details and step_details.type == "tool_calls":
            tool_call_detail = step_details.tool_calls
            for details in tool_call_detail:
                if isinstance(details, CodeInterpreterToolCall):
                    result = AnalysisResult(
                        tool_name = details.type,
                        agent_capability_category = AgentCapabilityCategories.OPENAI_ASSISTANTS
                    )
                    result.tool_input += details.code_interpreter.input  # Source code
                    for output in details.code_interpreter.outputs:  # Tool execution output
                        if hasattr(output, 'image') and output.image:
                            result.tool_output += "# Generated image file: " + output.image.file_id
                        elif hasattr(output, 'logs') and output.logs:
                            result.tool_output += output.logs
                    return result
                elif isinstance(details, FunctionToolCall):
                    result = AnalysisResult(
                        tool_name = details.function.name,
                        agent_capability_category = AgentCapabilityCategories.OPENAI_ASSISTANTS
                    )
                    result.tool_input += details.function.arguments
                    if details.function.output:
                        fn_output = json.loads(details.function.output)
                        if 'data' in fn_output:
                            output_data = json.loads(details.function.output)['data'][0]
                            result.tool_output += json.dumps({"url": output_data['url'], "description": output_data['revised_prompt']})
                        else:
                            # indicative of a failure during the function call, append error message to output
                            print("Error in function call: " + fn_output)
                            result.tool_output += json.dumps(fn_output)                            
                                
        return None

    @staticmethod
    def parse_message(message: Message):
        """
        Parses a message from the OpenAI Assistants API.

        Parameters
        ----------
        message : Message
            The message to parse.

        Returns
        -------
        List[MessageContentItemBase]
            The content items within the message along with any annotations.
        """
        ret_content = []
        # for each content item in the message
        for ci in message.content:
                match ci:
                    case TextContentBlock():
                        text_ci = OpenAITextMessageContentItem(
                            value=ci.text.value,
                            agent_capability_category = AgentCapabilityCategories.OPENAI_ASSISTANTS
                        )
                        for annotation in ci.text.annotations:
                            match annotation:
                                case FilePathAnnotation():
                                    file_an = OpenAIFilePathMessageContentItem(
                                        file_id=annotation.file_path.file_id,
                                        start_index=annotation.start_index,
                                        end_index=annotation.end_index,
                                        text=annotation.text,
                                        agent_capability_category = AgentCapabilityCategories.OPENAI_ASSISTANTS
                                    )
                                    text_ci.annotations.append(file_an)
                                case FileCitationAnnotation():
                                    file_cit = OpenAIFilePathMessageContentItem(
                                        file_id=annotation.file_citation.file_id,
                                        start_index=annotation.start_index,
                                        end_index=annotation.end_index,
                                        text=annotation.text,
                                        agent_capability_category = AgentCapabilityCategories.OPENAI_ASSISTANTS
                                    )
                                    text_ci.annotations.append(file_cit)
                        ret_content.append(text_ci)
                    case ImageFileContentBlock():
                        ci_img = OpenAIImageFileMessageContentItem(
                            file_id=ci.image_file.file_id,
                            agent_capability_category = AgentCapabilityCategories.OPENAI_ASSISTANTS
                        )
                        ret_content.append(ci_img)
                    case ImageURLContentBlock():
                        ci_img_url = OpenAIImageFileMessageContentItem(
                            file_url=ci.image_url.url,
                            agent_capability_category = AgentCapabilityCategories.OPENAI_ASSISTANTS
                        )
                        ret_content.append(ci_img_url)
        return ret_content



from foundationallm.models.resource_providers.ai_models import AIModelBase, AIModelTypes

class EmbeddingAIModel(AIModelBase):
    """
    An embedding AI model.
    """
    type: str = AIModelTypes.EMBEDDING



from pydantic import BaseModel, Field
from typing import Dict, List
from .resource_object_id_properties import ResourceObjectIdProperties

class ResourceObjectIdsModelBase(BaseModel):
    """
    The base model class for all models that contain a dictionary of resource object identifier properties.
    """

    resource_object_ids: Dict[str, ResourceObjectIdProperties] = Field(default_factory=dict, alias="resource_object_ids")

    def get_resource_object_id_properties(
            self,
            resource_provider_name: str,
            resource_type_name: str,
            property_name: str,
            property_value: str) -> ResourceObjectIdProperties:

        """
        Gets the resource object identifier properties for a specific resource provider, resource type, and object role.

        Args:
            resource_provider_name (str): The resource provider name.
            resource_type_name (str): The resource type name.
            object_role (str): The object role.

        Returns:
            ResourceObjectIdProperties: The resource object identifier properties.
        """
        return next(
            (v for v in self.resource_object_ids.values() \
                if v.resource_path.resource_provider == resource_provider_name \
                    and v.resource_path.main_resource_type == resource_type_name \
                        and property_name in v.properties \
                            and v.properties[property_name] == property_value), None)

    def get_many_resource_object_id_properties(
        self,
        resource_provider_name: str,
        resource_type_name: str,
        property_name: str,
        property_value: str) -> List[ResourceObjectIdProperties]:

        """
        Gets the resource object identifier properties for a specific resource provider, resource type, and object role.

        Args:
            resource_provider_name (str): The resource provider name.
            resource_type_name (str): The resource type name.
            object_role (str): The object role.

        Returns:
            List[ResourceObjectIdProperties]: A list of resource object identifier properties.
        """
        return [
            v for v in self.resource_object_ids.values() \
                if v.resource_path.resource_provider == resource_provider_name \
                    and v.resource_path.main_resource_type == resource_type_name \
                        and property_name in v.properties \
                            and v.properties[property_name] == property_value
        ]



"""
Class: TextEmbeddingRequest
Description:  Class representing a request into the gateway text embedding service.
"""
from pydantic import BaseModel
from typing import List
from .text_chunk import TextChunk

class TextEmbeddingRequest(BaseModel):
    """
    Class representing a request into the gateway text embedding service.
    """
    text_chunks: List[TextChunk]
    embedding_model_name: str
    prioritized: bool = False
    embedding_model_dimensions: int = 1536



import json
import time
from openai import AsyncAssistantEventHandler, AsyncAzureOpenAI
from openai.types.beta import AssistantStreamEvent
from openai.types.beta.threads import Text, TextDelta
from openai.types.beta.threads.runs import RunStep, RunStepDelta
from typing_extensions import override
from foundationallm.models.orchestration import CompletionResponse
from foundationallm.models.services import OpenAIAssistantsAPIRequest
from foundationallm.operations import OperationsManager
from foundationallm.services import ImageService
from foundationallm.utils import OpenAIAssistantsHelpers

class OpenAIAssistantAsyncEventHandler(AsyncAssistantEventHandler):
    """
    Event handler for streaming asynchronous OpenAI Assistant events.
    """
    def __init__(self, client: AsyncAzureOpenAI, operations_manager: OperationsManager, request: OpenAIAssistantsAPIRequest, image_service: ImageService):
        super().__init__()
        self.operations_manager = operations_manager
        self.request = request
        self.run_steps = {}
        self.messages = {}
        self.stop_tokens = [".", ",", ":", ";", "\n", " ", ")"] # Use stop tokens to determine when to write to State API.
        self.assistant_id = ''
        self.thread_id = ''
        self.interim_result = CompletionResponse(
            id = request.document_id,
            operation_id = request.operation_id,
            user_prompt = request.user_prompt,
            user_prompt_rewrite = request.user_prompt_rewrite,
            content = [],
            analysis_results = []
        )
        self.image_service = image_service
        self.client = client

    @override
    async def on_event(self, event: AssistantStreamEvent) -> None:
        if event.event == "thread.run.created":
            self.assistant_id = event.data.assistant_id
            self.thread_id = event.data.thread_id
        elif event.event == "thread.run.requires_action":
            await self.on_requires_action(event.data.id)
        elif event.event == "thread.run.step.created":
            self.run_steps[event.data.id] = event.data
        elif event.event == "thread.run.step.completed":
            details = event.data.step_details
            if details and details.type == "tool_calls":
                for tool in details.tool_calls or []:
                    if tool.type == "code_interpreter" and tool.code_interpreter and tool.code_interpreter.input and tool.code_interpreter.input.endswith(tuple(self.stop_tokens)):
                        self.run_steps[event.data.id] = event.data # Overwrite the run step with the final version.
                        await self.update_state_api_analysis_results_async()
                    if tool.type ==  "function":
                        self.run_steps[event.data.id] = event.data
                        await self.update_state_api_analysis_results_async()
        elif event.event == "thread.message.created":
            self.messages[event.data.id] = event.data
        elif event.event == "thread.message.completed":
            self.messages[event.data.id] = event.data # Overwrite the message with the final version.
            await self.update_state_api_content_async()
        elif "failed" in event.event:
            print(f'{event.event} ({event.data.id}): {event.data.last_error}')
            if event.event == "thread.run.failed":
                raise Exception(event.data.last_error.message)


    @override
    async def on_text_delta(self, delta: TextDelta, snapshot: Text) -> None:
        if snapshot.value.endswith(tuple(self.stop_tokens)): # Use stop tokens to determine when to write to State API.
            await self.update_state_api_content_async()

    @override
    async def on_run_step_delta(self, delta: RunStepDelta, snapshot: RunStep) -> None:
        details = delta.step_details
        if details and details.type == "tool_calls":
            for tool in details.tool_calls or []:
                if tool.type == "code_interpreter" and tool.code_interpreter and tool.code_interpreter.input and tool.code_interpreter.input.endswith(tuple(self.stop_tokens)):
                    self.run_steps[snapshot.id] = snapshot
                    await self.update_state_api_analysis_results_async()

    async def on_requires_action(self, run_id: str):
        max_steps = 15
        count = 0
        while count < max_steps:
            run = await self.client.beta.threads.runs.retrieve(run_id = run_id, thread_id = self.thread_id)
            count += 1
            print(f'\nPoll {count} run status: {run.status}')
            if run.status == "requires_action":
                tool_responses = []
                if (run.required_action.type == "submit_tool_outputs" and run.required_action.submit_tool_outputs.tool_calls is not None):
                        # Loop through each tool in the required action section
                        for call in run.required_action.submit_tool_outputs.tool_calls:
                            # Get data from the tool
                            if call.type == "function":
                                if call.function.name == "generate_image":
                                    try:
                                        tool_response = await self.image_service.generate_image_async(**json.loads(call.function.arguments))
                                        tool_responses.append(
                                            {
                                                "tool_call_id": call.id,
                                                "output": json.dumps(tool_response)
                                            }
                                        )
                                    except Exception as ex:
                                        print(f'Error getting tool response: {ex}')
                                        break
                try:
                    await self.client.beta.threads.runs.submit_tool_outputs(
                        run_id = run.id,
                        thread_id = self.thread_id,
                        tool_outputs = tool_responses,
                        stream = True
                    )
                except Exception as e:
                    print(f'Error submitting tool outputs: {e}')
                    continue
            if run.status == "failed":
                print('Run failed')
                break
            if run.status == "completed":
                break
            time.sleep(2)

    async def update_state_api_analysis_results_async(self):
        self.interim_result.analysis_results = [] # Clear the analysis results list before adding new results.
        for k, v in self.run_steps.items():
            if not v:
                continue
            analysis_result = OpenAIAssistantsHelpers.parse_run_step(v)
            if analysis_result:
                self.interim_result.analysis_results.append(analysis_result)
        await self.operations_manager.set_operation_result_async(self.request.operation_id, self.request.instance_id, self.interim_result)

    async def update_state_api_content_async(self):
        self.interim_result.content = [] # Clear the content list before adding new messages.
        for k, v in self.messages.items():
            content_items = OpenAIAssistantsHelpers.parse_message(v)
            self.interim_result.content.extend(content_items)

        await self.operations_manager.set_operation_result_async(self.request.operation_id, self.request.instance_id, self.interim_result)
